// If you have ever wanted to display only part of an image without having it displayed 
// as a rectangle, such as trees, plants, particles and interfaces, this tutorial
// will demonstrate this important technique.  You may have experimented or seen this
// accomplished with blending.  For example:  glBlendFunc(GL_SRC_ALPHA, GL_ONE) with
// blending turned on (glEnable(GL_BLEND)) will take a texture map and blend it with
// the background.  The closer a color is in the texture map to black, the more it
// will disappear.  If the color being blended is black, it will render the color 
// behind it instead.  This means you can have a picture of a tree with a black
// background and it will only draw the tree and not the black background surrounding it.
// The problem with this effect is that it makes the tree look transparent/faded.  This 
// works fine for particle effects, but if you want the image to be clear and not faded, 
// you will want to either use a 32 bit image with an alpha channel or masking.  Masking
// is basically making it so you only draw the part of an image where it's mask image
// has dark colors.  This means you have 2 images, one for the normal picture, and
// another black and white image that holds the mask.  More of this will be explained
// in the code and in the * QUICK NOTES * section at the bottom of this file. 

// 
// To demonstrate masking we will render a scope tunnel with a cross hair.  In between
// the end of the scope we will see our room rendered.  This is where the transparency
// will be.  If you look at the mask bitmap for the cross hair, you will see that this
// area is white, which means it will NOT draw that part of the image when combined
// with the original image.  Only the parts with dark colors will be drawn.  Let's
// go over the controls for this app:
//
// LEFT / RIGHT Arrow Keys  - This rotates the camera's view in a circle
// UP / DOWN Arrow Keys  - This makes our scope zoom in and out
// SPACE BAR - This turns night vision on and off
// ESCAPE Key - This quits the program
//
// Besides teaching masking, Ortho mode is also discussed.  Ortho mode allows us
// to use 2D coordinates instead of 3D coordinates.  This is perfect for art that
// needs to stay in front of the screen and never move, such as interface art.
// This is what we do for our scope image, this way it will always be centered on screen.
//
#include <iostream>
#include <math.h>
#include <GLUT/glut.h>
#include <stdlib.h>
#include "tgaClass.h"


#define MAX_TEXTURES	4
#define SPEED	0.03f								// This is how fast our camera moves

// Since we are using a full screen picture for the scope, we should bring down our
// resolution so we don't have a much larger file in texture memory.
#define SCREEN_WIDTH 640						// We want our screen width 640 pixels
#define SCREEN_HEIGHT 480

// This will store whether we want night vision on our off (press the space bar)
bool g_bNightVision = false;

// This is our camera rotation degree
float g_rotateY = 0;

// This holds the zoom value of our scope
float g_zoom = 0;
	

using namespace std;

void display();
void init();
void reshape(int w, int h);
void keyboard(unsigned char key, int x, int y);
void arrowkeys(int key, int x, int y);
void CreateTexture(unsigned int textureArray[], char * strFileName, int textureID);
void PerspectiveMode();
void OrthoMode(int left, int top, int right, int bottom);
void DrawRoom();

unsigned int g_Texture[MAX_TEXTURES] = {0};
	
	
///////////////////////////////// CREATE TEXTURE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This creates a texture in OpenGL that we can texture map
/////
///////////////////////////////// CREATE TEXTURE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void CreateTexture(unsigned int textureArray[], char * strFileName, int textureID)
{
	
	char buffer[30];	//filename holder
	int status = 0;		//error codes for file read
	TGA myTGAImage;
	
	if(!strFileName){			// Return from the function if no file name was passed in
		
		cout <<"Error occurred, bad file name "<< endl;
		return;
	}
	
	sprintf (buffer, strFileName);		//load buffer with filename
	status = myTGAImage.readTGA(buffer);// Load into a TGA object
	if(status > 1) {
		cout <<"Error occurred = " << status <<strFileName<< endl;
		exit(0);
	}


	// Now that we have the texture data, we need to register our texture with OpenGL
	// To do this we need to call glGenTextures().  The 1 for the first parameter is
	// how many texture we want to register this time (we could do a bunch in a row).
	// The second parameter is the array index that will hold the reference to this texture.

	// Generate a texture with the associative texture ID stored in the array
	glGenTextures((GLsizei)1, (GLuint *)&textureArray[textureID]);

	// This sets the alignment requirements for the start of each pixel row in memory.
	glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
	
	// Now that we have a reference for the texture, we need to bind the texture
	// to tell OpenGL this is the reference that we are assigning the bitmap data too.
	// The first parameter tells OpenGL we want are using a 2D texture, while the
	// second parameter passes in the reference we are going to assign the texture too.
	// We will use this function later to tell OpenGL we want to use this texture to texture map.

	// Bind the texture to the texture arrays index and init the texture
	glBindTexture(GL_TEXTURE_2D, textureArray[textureID]);

	// Now comes the important part, we actually pass in all the data from the bitmap to
	// create the texture. Here is what the parameters mean in gluBuild2DMipmaps():
	// (We want a 2D texture, 3 channels (RGB), bitmap width, bitmap height, It's an RGB format,
	//  the data is stored as unsigned bytes, and the actuall pixel data);

	// What is a Mip map?  Mip maps are a bunch of scaled pictures from the original.  This makes
	// it look better when we are near and farther away from the texture map.  It chooses the
	// best looking scaled size depending on where the camera is according to the texture map.
	// Otherwise, if we didn't use mip maps, it would scale the original UP and down which would
	// look not so good when we got far away or up close, it would look pixelated.

	// Build Mipmaps (builds different versions of the picture for distances - looks better)
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3,  myTGAImage.width, 
					  myTGAImage.height, myTGAImage.format, GL_UNSIGNED_BYTE, myTGAImage.data);


	// Lastly, we need to tell OpenGL the quality of our texture map.  GL_LINEAR is the smoothest.
	// GL_NEAREST is faster than GL_LINEAR, but looks blochy and pixelated.  Good for slower computers though.
	// Read more about the MIN and MAG filters at the bottom of main.cpp
		
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	
}

//Display callback
void display()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear The Screen And The Depth Buffer
	glLoadIdentity();									// Reset The matrix
	
	/////// * /////////// * /////////// * NEW * /////// * /////////// * /////////// *
// Instead of adding our camera class to this tutorial, I just did a quick hack to
	// demonstrate the zooming in and out of the scope.  Ideally, we would have a position
	// for the camera and a separate camera object for the scope.  We just need to zoom
	// in along the Z axis, because g_rotateY takes care of the rotations.

		// 	  Position      		View	   Up Vector
	gluLookAt(0, 0, g_zoom,     0, 0, g_zoom + 5,     0, 1, 0);	

	// Rotate the world around the y axis, according to the degree stored in g_rotateY
	glRotatef(g_rotateY , 0, 1, 0);
	
	// If g_bNightVision is true, we want to turn our world green for night vision,
	// otherwise do the normal color.  glColor4f() takes a RGB and alpha value,
	if(g_bNightVision)
		glColor4f(0, 1, 0, 1);
	else
		glColor4f(1, 1, 1, 1);
				
	// Here we draw the texture walls and floor
	DrawRoom();

	// Before drawing the scope texture, we need to switch to ortho (2D) mode.  We pass in
	// our screen width and height to use as the right and bottom coordinates of the view port.
	OrthoMode(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	
	
	///////////////////// MASKING ///////////////////////////////

	// Once the room is drawn, we need to start doing the masking.  In order for
	// this to work, we NEED TO TURN OFF DEPTH TESTING.  Otherwise it won't work.
	// Before we render the real image, we will render the mask image.


	// Select the desired blending function and turn on blending; Refer to the Blending discussion in the RedBook
        // starting on page 162. Notes below ->
        //In this case we want the source (the mask) colors to be multiplied by the contents of the frame buffer (destination). So the 	
        //source blending factor is determined by the color stored in the buffer. 
        // Furthermore, we do not want the frame buffer (destination) colors to be added to the source values. Which blending constant
        // should be used for the destination?
        //*****Add your code here********//											

	// Bind the mask texture to our new 2D quad
	glBindTexture(GL_TEXTURE_2D,  g_Texture[2]);
	
	// Display a 2D quad with the scope/cross hair mask. The mask is textured over the entire screen space.
	glBegin(GL_QUADS);
	
		// Notice that when we are in ortho mode, we use glVertex2f() to pass
		// in screen coordinates, not vertices.  This makes it incredibly easy
		// to put up 2D interface art.  It's just like doing 2D graphics.
		// The texture coordinate stay the same regardless.
                //You are required to determine the 2D vertices that define the quad for the cross hair mask

		// Display the top left point of the 2D image
		glTexCoord2f(0.0f, 1.0f);	//*****Add your code here********//

		// Display the bottom left point of the 2D image
		glTexCoord2f(0.0f, 0.0f);	//*****Add your code here********//

		// Display the bottom right point of the 2D image
		glTexCoord2f(1.0f, 0.0f);	//*****Add your code here********//

		// Display the top right point of the 2D image
		glTexCoord2f(1.0f, 1.0f);	//*****Add your code here********//

	// Stop drawing 
	glEnd();	
	
	// Once the mask is draw to the screen, we want to set a new blend mode.
	// This mode will perform the masking effect that we desire.  Then we just
	// render the desired transparent image over top of the mask and it will
	// only draw the desired effect. Here we want the source and destination colors to be added without modification. Which
        //multiplier (blending factor) gives us this effect?
        // Select the desired blending function
        //*****Add your code here********//
        
                	
	// Bind the scope texture to this next 2D quad
	glBindTexture(GL_TEXTURE_2D,  g_Texture[3]);

	// Display a 2D quad with our scope texture
	glBegin(GL_QUADS);
                //You are required to determine the 2D vertices that define the quad for the cross hair scope
		// Display the top left point of the 2D image
		glTexCoord2f(0.0f, 1.0f);	//*****Add your code here********//

		// Display the bottom left point of the 2D image
		glTexCoord2f(0.0f, 0.0f);	//*****Add your code here********//

		// Display the bottom right point of the 2D image
		glTexCoord2f(1.0f, 0.0f);	//*****Add your code here********//

		// Display the top right point of the 2D image
		glTexCoord2f(1.0f, 1.0f);	//*****Add your code here********//

	// Stop drawing 
	
	glEnd();

	// Since we are done, we can now turn OFF blending and turn ON depth testing
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);

	// Let's set our mode back to perspective 3D mode.  None of this archaic 2D stuff :)
	PerspectiveMode();  

	glutSwapBuffers();
}

// Here comes the complicated stuff that we don't usually deal with, our projection matrix.
// Basically what is going on with the 2 functions OrthoMode() and PerspectiveMode() is
// we are switching from plotting 3D point to plotting 2D points.  This means that
// instead of giving OpenGL 3D coordinates, in Ortho mode we give it screen pixel coords.
// This is a really useful mode to have because it saves us tons of math when it comes
// to displaying interface art over our 3D screen.  You would have to rotate the
// interface art whenever the camera changes, which could be annoying.  On top of that,
// you need to take into account the depth value, which means you have to find the best
// depth to render the art at.  This can stretch and pixelate the art, making it look bad.
// This is also incredibly faster than doing glDrawPixels(), and you don't have to touch
// the rasterizer, plus you can use blending and other OpenGL effects.

///////////////////////////////// ORTHO MODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function changes our projection mode from 3D to 2D
/////
///////////////////////////////// ORTHO MODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void OrthoMode(int left, int top, int right, int bottom)
{
	// This function takes in rectangle coordinates for our 2D view port.
	// What happens below is, we need to change our projection matrix.  To do this,
	// we call glMatrixMode(GL_PROJECTION) to tell OpenGL the current mode we want.
	// Then we push on a new matrix so we can just call glPopMatrix() to return us
	// to our previous projection matrix, but first we will render in ortho mode.
	//  Next, we load a new identity matrix so that everything is initialize before
	// we go into ortho mode.  One we switch to ortho by calling glOrth(), we then
	// want to set our matrix mode to GL_MODELVIEW, which puts us back and ready to
	// render our world using the model matrix.  We also initialize this new model view
	// matrix for our 2D rendering before drawing anything.

	// Switch to our projection matrix so that we can change it to ortho mode, not perspective.
	glMatrixMode(GL_PROJECTION);						

	// Push on a new matrix so that we can just pop it off to go back to perspective mode
	glPushMatrix();
	
	// Reset the current matrix to our identify matrix
	glLoadIdentity();								

	//Pass in our 2D ortho screen coordinates.like so (left, right, bottom, top).  The last
	// 2 parameters are the near and far planes.
	glOrtho( left, right, bottom, top, 0, 1 );	
	
	// Switch to model view so that we can render the scope image
	glMatrixMode(GL_MODELVIEW);								

	// Initialize the current model view matrix with the identity matrix
	glLoadIdentity();										
}


///////////////////////////////// PERSPECTIVE MODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function changes our returns our projection mode from 2D to 3D
/////
///////////////////////////////// PERSPECTIVE MODE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
void PerspectiveMode()										// Set Up A Perspective View
{
	// This function doesn't really go to perspective mode, since we aren't
	// using gluPerspective(), but if you look in OrthoMode(), we push on
	// a matrix once we change to our projection matrix.  All we need to do is
	// pop that matrix off of our stack and it will return to our original perspective
	// matrix.  So, before we pop the current matrix, we need to tell OpenGL which
	// matrix we want to use by changing our current matrix mode to perspective.
	// Then we can go back to our model view matrix and render normally.

	// Enter into our projection matrix mode
	glMatrixMode( GL_PROJECTION );							

	// Pop off the last matrix pushed on when in projection mode (Get rid of ortho mode)
	glPopMatrix();											

	// Go back to our model view matrix like normal
	glMatrixMode( GL_MODELVIEW );							

	// We should be in the normal 3D perspective mode now
}

///////////////////////////////// DRAW ROOM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function renders the room walls and floor
/////
///////////////////////////////// DRAW ROOM \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void DrawRoom()
{
	// To get a better view of our scope transparency, create some walls and
	// a floor to look at.  Use the wall texture and floor texture for the
	// quads that are rendered.  The camera is in the center
	// of the world at the origin, and the walls are 15 units away from the camera. The 
        // height of a wall is 2 units, centered at the origin. So, each wall is 15 X 15 X 2.
	// Since the quads are so huge, we want to wrap the seamless texture multiple
	// times instead of stretching the texture along the huge wall.  So the
	// S (x value) for the ST coordinates is greater than 1 for the right-edge of the texture map.
        // If you use (2,0) and (2,1) as the coordinates for the right-edge of the texture then the texture is repeated
        // twice across the face of the quad.  12 repetitions is
	// sufficient. This is the most efficient way to wrap (repeat a texture) - defining
        // multiple quads and texturing each quad is a waste. 

	// Bind the wall texture to the wall quads
	glBindTexture(GL_TEXTURE_2D,  g_Texture[0]);

	// Display 4 long quads to represent walls
	glBegin(GL_QUADS);

	////////////////////// BACK WALL ///////////////////////

		//*****Add your code here********//

	////////////////////// FRONT WALL ///////////////////////

		//*****Add your code here********//

	////////////////////// LEFT WALL ///////////////////////
		
                //*****Add your code here********//


	////////////////////// RIGHT WALL ///////////////////////

		//*****Add your code here********//

	// Stop drawing the walls
	glEnd();
	
	// Now we need to draw the floor.  Wrap the seamless texture
	// along the floor 16 times each way.  See if you can figure out the texture
        //coordinates for this one.
		
	// Bind this texture to the floor quad
	glBindTexture(GL_TEXTURE_2D,  g_Texture[1]);

	// Display a huge quad to represent a floor
	glBegin(GL_QUADS);

		//*****Add your code here********//

	// Stop drawing the floor
	glEnd();
}


void keyboard(unsigned char key, int x, int y)
{
#pragma unused(x,y)
   switch (key) {
      case 27:
      	int w = glutGetWindow();
      	glutDestroyWindow(w);
      	glDeleteTextures((GLsizei)MAX_TEXTURES, (GLuint *)g_Texture);
        exit(0);  /*  Escape key  */
        break;
      case ' ':
      	g_bNightVision = !g_bNightVision;
      	glutPostRedisplay();
      	break;
      default:
         break;
   }
}

void arrowkeys(int key, int x, int y)
{
#pragma unused(x,y)
   switch (key) {
      case GLUT_KEY_UP:								// If we hit the UP arrow key
		// Make sure we haven't gone past or max scope range, if not, increase our zoom level
		if(g_zoom < 13) g_zoom += 0.5f;		
		glutPostRedisplay();						// Redraw the screen to reflect changes
		break;
	  case GLUT_KEY_DOWN:							// If we hit the DOWN arrow key
		// Make sure we don't zoom backwards, if we haven't go below 0, decrease our zoom level
		if(g_zoom > 0)  g_zoom -= 0.5f;			
		glutPostRedisplay();						// Redraw the screen to reflect changes
		break;
	  case GLUT_KEY_LEFT:							// If we hit the DOWN arrow key
		g_rotateY -= 2;								// Rotate camera left
		glutPostRedisplay();						// Redraw the screen to reflect changes
		break;
	  case GLUT_KEY_RIGHT:							// If we hit the DOWN arrow key
		g_rotateY += 2;								// Rotate camera right
		glutPostRedisplay();						// Redraw the screen to reflect changes
		break; 
      default:
         break;
   }
  }

//Init callback
void init()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	int width = glutGet(GLUT_WINDOW_WIDTH);
	int height = glutGet(GLUT_WINDOW_HEIGHT);
	glViewport(0,0,width,height);
	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height, .5f, 150.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Here we load all of our textures. 
	CreateTexture(g_Texture, "Wall.tga", 0);			// Load our texture for the wall
	CreateTexture(g_Texture, "floor.tga", 1);			// Load our floor texture
	CreateTexture(g_Texture, "CrossHairMask.tga", 2);	// Load the cross hair mask used for transparency
	CreateTexture(g_Texture, "CrossHair.tga", 3);	    // Load the scope/cross hair image
	
	// Instead of boring black for the clear color, I wanted to make it a nice sky blue.
	// The parameters are (Red Value, Green Value, Blue Value, Alpha Value).  Now our
	// background will have a blue color, instead of clearing away to black.
	glClearColor(.4f, 0, 1, 1);

	glEnable(GL_TEXTURE_2D);							// Enable Texture Mapping
}

int main(int argc, char **argv)
{
	
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("Masking");
	glutFullScreen();
	glutDisplayFunc(display);
	glutSpecialFunc(arrowkeys);
	glutKeyboardFunc(keyboard);
	init();
	glutMainLoop();
}